/*
 * authkeyprov.c - authentication key provider interface
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef HAVE_DLFCN_H
# include <dlfcn.h>
#endif

#include <stdio.h>

#include "ntp_types.h"
#include "ntp_fp.h"
#include "ntp.h"
#include "ntpd.h"
#include "ntp_string.h"
#include "ntp_malloc.h"
#include "ntp_stdlib.h"
#include "ntp_authprov.h"
#include "ntp_request.h"

/*
 * Allow dynamic linker flags to be overriden at configure time
 * for very special cases.
 */
#ifndef NTP_AUTHPROV_RTLD_FLAGS
#define NTP_AUTHPROV_RTLD_FLAGS	RTLD_LOCAL
#endif

struct authprov {
	void *handle;
	struct authkey_be *be;
	keyid_t bounds[2];
	struct authprov *next;
};

static struct authprov *authprovs = NULL;

/*
 * authprov_select - select authentication provider backend for a given key ID
 */
struct authkey_be *
authprov_select(keyid_t keyno)
{
	struct authprov *ap;

	if (keyno == 0) {
		return (NULL);
	}

	for (ap = authprovs; ap != NULL; ap = ap->next) {
		if (keyno >= ap->bounds[0] && keyno <= ap->bounds[1]) {
			break;
		}
	}

	if (ap == NULL) {
		return (NULL);
	}

	return (ap->be);
}

/*
 * init_auth - initialize default authentication provider list
 *
 * Note that builtin authentication providers can be explicitly
 * instantiated with the "authprov" configuration file directive,
 * which allows (for example) a different key file to be used
 * for a portion of the key ID space. The default providers are
 * always registered, but it is possible for the key ID space
 * they manage to be completely obscured by user providers.
 */
int
init_auth(const char *keyfile)
{
	struct authkey_be *be;
	keyid_t bounds[2];

	/*
	 * Register keyfile authentication provider for key ID
	 * space between 0 and NTP_MAXKEY.
	 */
	be = keyfile_authprov_init(AUTHPROV_VERSION, keyfile != NULL, &keyfile);
	if (be == NULL) {
		return (0);
	}

	bounds[0] = 0;
	bounds[1] = NTP_MAXKEY;

	if (!authprov_register(KEYFILE_AUTHPROV_BE, be, bounds, 0)) {
		return (0);
	}

	/*
	 * Register autokey authenetication provider for key ID
	 * space from NTP_MAXKEY + 1.
	 */
	be = autokey_authprov_init(AUTHPROV_VERSION, 0, NULL);
	if (be == NULL) {
		return (0);
	}

	bounds[0] = NTP_MAXKEY + 1;
	bounds[1] = (keyid_t)~0;

	if (!authprov_register(AUTOKEY_AUTHPROV_BE, be, bounds, 0)) {
		return (0);
	}

	return (1);
}

/*
 * authprov_genkeyid - generate a key ID within bounds for an authentication
 * provider. We can't call authprov_select() without a keyid, so instead
 * this function takes a handle (which is typically a well known handle).
 */
int
authprov_genkeyid(
	void *handle,
	keyid_t *pkeyid)
{
	struct authprov *ap;
	keyid_t keyid;

	for (ap = authprovs; ap != NULL; ap = ap->next) {
		if (ap->handle == handle) {
			break;
		}
	}

	if (ap == NULL) {
		return (0);
	}

	while (1) {
		keyid = (ntp_random() + ap->bounds[0]);
		if (keyid > ap->bounds[1])
			keyid = ap->bounds[1];
		/* This check is necessary in case of overlapping bounds */
		if (authprov_select(keyid) != ap->be)
			continue;
		if (ap->be->havekey != NULL && (*ap->be->havekey)(ap->be, keyid, 1))
			continue;
		break;
	}

	*pkeyid = keyid;

	return (1);
}

/*
 * authprov_register - register a new authentication provider
 */
int
authprov_register(
	void *handle,
	struct authkey_be *be,
	keyid_t bounds[2],
	int userprov
	)
{
	struct authprov *ap, **app;

	/*
	 * Each authentication provider is assigned a portion of the keyid
	 * space. For maximum flexibility, the space is configured by the
	 * administrator rather than the provider itself.
	 *
	 * In the default case where no user providers are registered, the
	 * NTP_MAXKEY pivot is preserved.
	 */
	for (app = &authprovs; *app != NULL; app = &(*app)->next) {
		ap = *app;

		/*
		 * First registered provider wins in case of bounds overlap.
		 * Internal providers are registered after user providers.
		 *
		 * Only print a warning if the overlap is between user providers.
		 */
		if (userprov &&
		    (bounds[0] >= ap->bounds[0] || bounds[1] <= ap->bounds[1])) {
			fprintf(stderr, "Warning: authentication provider bounds %u:%u "
				"overlap existing bounds %u:%u\n",
				bounds[0], bounds[1], ap->bounds[0], ap->bounds[1]);
		}
	}

	ap = (struct authprov *)calloc(1, sizeof(*ap));
	if (ap == NULL) {
		return (0);
	}	

	ap->handle = handle;
	ap->be = be;
	ap->bounds[0] = bounds[0];
	ap->bounds[1] = bounds[1];
	ap->next = NULL;

	*app = ap;

	return (1);
}

/*
 * authprov_destroy - cleanup a single authentication provider
 */
static void
authprov_destroy(void *handle, struct authkey_be *be)
{
	if (be->destructor != NULL) {
		(*be->destructor)(be);
	}
#ifdef HAVE_DLFCN_H
	if (handle != KEYFILE_AUTHPROV_BE && handle != AUTOKEY_AUTHPROV_BE) {
		dlclose(handle);
	}
#endif /* HAVE_DLFCN_H */
}

/* 
 * authprov_cleanup - cleanup registered authentication providers
 */
void
authprov_cleanup(void)
{
	struct authprov *ap, *next;

	for (ap = authprovs; ap != NULL; ap = next) {
		next = ap->next;
		authprov_destroy(ap->handle, ap->be);
		memset(ap, 0, sizeof(*ap));
		free(ap);
	}
}

/*
 * authprov_refresh - refresh all authentication providers
 */
void
authprov_refresh(void)
{
	struct authprov *ap;

	for (ap = authprovs; ap != NULL; ap = ap->next) {
		if (ap->be->refresh != NULL) {
			(*ap->be->refresh)(ap->be);
		}
	}
}

/*
 * authprov_load - load a new authentication provider
 *
 * Path is either a path to a dynamically loadable module (for platforms
 * that support it) or one of the well known constants "keyfile" or
 * "autokey"
 */
int
authprov_load(
	const char *path,
	keyid_t bounds[2],
	int argc,
	const char *argv[]
	)
{
	struct authkey_be *be;
	void *handle;
	const char *err;
	authkey_be_constructor constr;

	if (strcmp(path, "keyfile") == 0) {
		constr = keyfile_authprov_init;
		handle = KEYFILE_AUTHPROV_BE;
	} else if (strcmp(path, "autokey") == 0) {
		constr = autokey_authprov_init;
		handle = AUTOKEY_AUTHPROV_BE;
	} else {
#ifdef HAVE_DLFCN_H
		handle = dlopen(path, RTLD_NOW | NTP_AUTHPROV_RTLD_FLAGS);
		if (handle == NULL) {
			err = dlerror();
			fprintf(stderr, "Failed to load library %s: %s\n", path,
				err != NULL ? err : strerror(errno));
			return (0);
		}

		dlerror();

		/*
		 * Each provider should export a symbol authprov_init, being a
		 * constructor that returns an authkey_be structure.
		 */
		constr = (authkey_be_constructor)dlsym(handle, "authprov_init");
		if (constr == NULL) {
			err = dlerror();
			fprintf(stderr, "Failed to resolve symbol authprov_init: %s",
				err != NULL ? err : strerror(errno));
			dlclose(handle);
			return (0);
		}
#else
		fprintf(stderr, "Loadable authentication providers not supported on this platform\n");
		return (0);
#endif /* HAVE_DLFCN_H */
	}

	be = (*constr)(AUTHPROV_VERSION_V1, argc, argv);
	if (be == NULL) {
		fprintf(stderr, "Failed to initialize NTP authentication provider %s\n",
			path);
		authprov_destroy(handle, be);
		return (0);
	}

	if (!authprov_register(handle, be, bounds, 1)) {
		fprintf(stderr, "Failed to register NTP authentication provider %s\n",
			path);
		authprov_destroy(handle, be);
		return (0);
	}

	return (1);
}

/*
 * authsetkey - associate a key with a key ID.
 *
 * If the trust argument is 0, the key is untrusted; if 1, the key is
 * trusted; any other value denotes the trust lifetime in seconds.
 */
int
authsetkey(
	keyid_t keyno,
	int type, /* should be KEY_TYPE_MD5 for now */
	const u_char *key,
	const int len,
	u_long trust
	)
{
	struct authkey_be *be;

	be = authprov_select(keyno);
	if (be == NULL) {
		return (0);
	}

	if (be->setkey == NULL) {
		return (0);
	}

	return (*be->setkey)(be, keyno, type, key, len, trust);
}

/*
 * auth_havekey - returns one if the key is known
 */
int
auth_havekey(
	keyid_t keyno
	)
{
	struct authkey_be *be;

	be = authprov_select(keyno);
	if (be == NULL) {
		return (0);
	}

	if (be->havekey == NULL) {
		return (0);
	}

	return (*be->havekey)(be, keyno, 0);
}

/*
 * authhavekey - return one and cache the key, if known and trusted.
 */
int
authhavekey(
	keyid_t keyno
	)
{
	struct authkey_be *be;

	if (keyno == 0) {
		return (1); /* to mirror existing behaviour */
	}

	be = authprov_select(keyno);
	if (be == NULL) {
		return (0);
	}

	if (be->havekey == NULL) {
		return (0);
	}

	return (*be->havekey)(be, keyno, 1);
}

/*
 * authtrust - declare a key to be trusted/untrusted
 */
void
authtrust(
	keyid_t keyno,
	u_long trust
	)
{
	struct authkey_be *be;

	be = authprov_select(keyno);
	if (be == NULL) {
		return;
	}

	if (be->trust == NULL) {
		return;
	}

	(*be->trust)(be, keyno, trust);
}

/*
 * authistrusted - determine whether a key is trusted
 */
int
authistrusted(
	keyid_t keyno
	)
{
	struct authkey_be *be;

	be = authprov_select(keyno);
	if (be == NULL) {
		return (4);
	}

	if (be->istrusted == NULL) {
		return (0);
	}

	return (*be->istrusted)(be, keyno);
}

/*
 * auth_agekeys - delete keys whose lifetimes have expired
 */
void
auth_agekeys(void)
{
	struct authprov *ap;

	for (ap = authprovs; ap != NULL; ap = ap->next) {
		if (ap->be->agekeys != NULL) {
			(*ap->be->agekeys)(ap->be);
		}
	}
}

/*
 * authencrypt - generate message authenticator
 *
 * Returns length of authenticator field, zero if key not found.
 */
int
authencrypt(
	keyid_t keyno,
	u_int32 *pkt,
	int length
	)
{
	struct authkey_be *be;

	pkt[length / 4] = htonl(keyno);

	be = authprov_select(keyno);
	if (be == NULL) {
		return (4);
	}

	if (be->getmac == NULL) {
		return (0);
	}

	return (*be->getmac)(be, keyno, pkt, length);
}

/*
 * authdecrypt - verify message authenticator
 *
 * Returns one if authenticator valid, zero if invalid or key not found.
 */
int
authdecrypt(
	keyid_t keyno,
	u_int32 *pkt,
	int length,
	int size
	)
{
	struct authkey_be *be;

	be = authprov_select(keyno);
	if (be == NULL) {
		return (0);
	}

	if (be->verifymac == NULL) {
		return (0);
	}

	return (*be->verifymac)(be, keyno, pkt, length, size);
}

/*
 * authprov_stats - gather aggregate statistics for authentication providers
 */
int
authprov_stats(struct info_auth *ia)
{
	struct authprov *ap;
	struct info_auth total_stats;

	memset(&total_stats, 0, sizeof(total_stats));

	for (ap = authprovs; ap != NULL; ap = ap->next) {
		if (ap->be->stats != NULL) {
			struct info_auth stats;

			memset(&stats, 0, sizeof(stats));

			if (!(*ap->be->stats)(ap->be, &stats)) {
				continue;
			}

			total_stats.numkeys	+= stats.numkeys;
			total_stats.numfreekeys	+= stats.numfreekeys;
			total_stats.keylookups	+= stats.keylookups;
			total_stats.keynotfound	+= stats.keynotfound;
			total_stats.encryptions	+= stats.encryptions;
			total_stats.decryptions	+= stats.decryptions;
			total_stats.keyuncached	+= stats.keyuncached;
			total_stats.expired	+= stats.expired;
		}
	}

	ia->numkeys	= htonl((u_int32)total_stats.numkeys);
	ia->numfreekeys	= htonl((u_int32)total_stats.numfreekeys);
	ia->keylookups	= htonl((u_int32)total_stats.keylookups);
	ia->keynotfound	= htonl((u_int32)total_stats.keynotfound);
	ia->encryptions	= htonl((u_int32)total_stats.encryptions);
	ia->decryptions	= htonl((u_int32)total_stats.decryptions);
	ia->keyuncached	= htonl((u_int32)total_stats.keyuncached);
	ia->expired	= htonl((u_int32)total_stats.expired);

	return (1);
}

/*
 * authprov_reset_stats - reset authentication provider statistics
 */
void
authprov_reset_stats(void)
{
	struct authprov *ap;

	for (ap = authprovs; ap != NULL; ap = ap->next) {
		if (ap->be->reset_stats != NULL) {
			(*ap->be->reset_stats)(ap->be);
		}
	}
}

